# function subroutine to for sub pixel localization using MLE#
import numpy as np
from scipy.optimize import minimize
import numba
from joblib import Parallel, delayed

# Define the 2D Gaussian model
@numba.jit
def twoD_Gaussian(xdata_tuple, amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
    """
    Calculates the value of a 2D Gaussian function at given coordinates.
    """
    (x, y) = xdata_tuple
    xo = float(xo)
    yo = float(yo)
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    g = offset + amplitude * \
        np.exp(- (a*((x-xo)**2) + 2*b*(x-xo)*(y-yo) + c*((y-yo)**2)))
    return g

# Define the negative log-likelihood function (to be minimized)
def likelihood(params, image):
    """
    Likelihood function for the MLE algorithm.
    Calculates the negative log likelihood.
    """
    regularization = 1e-10				# a very small value, so that the log operation doesnot face any error
    
    amplitude, xo, yo, sigma_x, sigma_y, theta, offset = params

    x, y = np.meshgrid(np.arange(image.shape[1]), np.arange(image.shape[0]))

    expected_image = twoD_Gaussian((x, y), amplitude, xo, yo, sigma_x, sigma_y, theta, offset)
    
    # Apply regularization to avoid zero values
    expected_image = np.maximum(expected_image, regularization)  
	
	# Flatten both image and model arrays for efficient processing
    #image = image.ravel()				
    #expected_image = expected_image.ravel()

    
    log_likelihood = np.sum(image * np.log(expected_image) - expected_image)
    return -log_likelihood


# MLE
def localizer(coordinates, images, window_size, initial_guess, method):#, num_cores):
    """
    Localizes molecules using maximum likelihood estimation (MLE).
    """

    num_cores=10            # Number of CPU cores to use for parallel processing


    # Collect the x, y, and intensity values of all the molecules
    
    molecule_intensity = []
    localized_molecule_x = []
    localized_molecule_y = []
    sigma_x = []
    sigma_y = []
    

    params_fit = [[0]]*5

    if len(coordinates) > 0:
        if method == 'BlobDetection':
            x, y = coordinates[0], coordinates[1]
        else:
            x, y = coordinates

        def process_molecule(j):
            x1, y1 = int(x[0][j]), int(y[0][j])

            # Extract a sub-image around the molecule
            sub_image = images[y1-window_size:y1+window_size, x1-window_size:x1+window_size]
            
            

            #bounds=((0,255),(0,sub_image[1]),(0,sub_image[0]),(0,np.inf),(0,np.inf),(0,100),(0,2*np.pi))
            #bounds = [(0,255), (0,sub_image[1]), (0,sub_image[0]),(0,window_size), (0,window_size), (0,2*np.pi), (0,100)]
            try: 
                # Perform minimization of negative log-likelihood to estimate parameters
                popt = minimize(likelihood, initial_guess, args=(sub_image),method='Nelder-Mead', tol=1e-2)
                return [popt.x[0], (x1-window_size)+popt.x[1], (y1-window_size)+popt.x[2], popt.x[3], popt.x[4]]
            except:
                return None
        
        # Process molecules in parallel
        results = Parallel(n_jobs=num_cores)(delayed(process_molecule)(j) for j in range(len(x)))
        
        for result in results:
            if result is not None:
                # Collect the estimated parameters of the molecules
                molecule_intensity.append(result[0])
                localized_molecule_x.append(result[1])
                localized_molecule_y.append(result[2])
                sigma_x.append(result[3])
                sigma_y.append(result[4])

        params_fit = molecule_intensity, localized_molecule_x, localized_molecule_y, sigma_x, sigma_y

    return params_fit
